import { generate } from "shortid"
import {
  tempCodeExpiryLength,
  USERS_LOCK_FILE,
  USERS_LIST_FILE,
  userAuthFile,
  getUserByName,
} from "./authCommon"
import { getLock, isNolock, releaseLock } from "../common/lock"
import { apiWrapper, events } from "../common"
import { alwaysAuthorized } from "./permissions"

export const createTemporaryAccess = app => async userName =>
  apiWrapper(
    app,
    events.authApi.createTemporaryAccess,
    alwaysAuthorized,
    { userName },
    _createTemporaryAccess,
    app,
    userName
  )

export const _createTemporaryAccess = async (app, userName) => {
  const tempCode = await getTemporaryCode(app)

  const lock = await getLock(app, USERS_LOCK_FILE, 1000, 2)

  if (isNolock(lock)) {
    throw new Error(
      "Unable to create temporary access, could not get lock - try again"
    )
  }

  try {
    const users = await app.datastore.loadJson(USERS_LIST_FILE)

    const user = getUserByName(users, userName)
    user.temporaryAccessId = tempCode.temporaryAccessId

    await app.datastore.updateJson(USERS_LIST_FILE, users)
  } finally {
    await releaseLock(app, lock)
  }

  const userAuth = await app.datastore.loadJson(userAuthFile(userName))
  userAuth.temporaryAccessHash = tempCode.temporaryAccessHash

  userAuth.temporaryAccessExpiryEpoch = tempCode.temporaryAccessExpiryEpoch

  await app.datastore.updateJson(userAuthFile(userName), userAuth)

  return tempCode.tempCode
}

export const getTemporaryCode = async app => {
  const tempCode = generate() + generate() + generate() + generate()

  const tempId = generate()

  return {
    temporaryAccessHash: app.crypto.hash(tempCode),
    temporaryAccessExpiryEpoch:
      (await app.getEpochTime()) + tempCodeExpiryLength,
    tempCode: `tmp:${tempId}:${tempCode}`,
    temporaryAccessId: tempId,
  }
}

export const looksLikeTemporaryCode = code => code.startsWith("tmp:")
